/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sched.h>
#include <memory.h>
#include <pthread.h>
#include <stdio.h>
#include "bufferManager.h"
#include "page.h"
#include "cfs.h"

//----------------------------------------
// Constructor of the class BufMgr
//----------------------------------------

Page *BufMgr::lowFrameAddress;    // used to check that data pointer is within the frame buffer's space
Page *BufMgr::highFrameAddress;
ReadList *BufMgr::staticReadList;

void BufMgr::printReadList () {
  cout << *BufMgr::staticReadList << endl;
}

BufMgr::BufMgr(const int bufs, const int bufferSize, Status& returnStatus) : numBufs (bufs)
{
  //  numBufs = bufs;

  // eh.  wrong
  //  readList = new ReadList (numBufs);
  int readListSize = numBufs;
  if (CFS::groupCount > readListSize)
    readListSize = CFS::groupCount;

  readList = new ReadList (readListSize);

  BufMgr::staticReadList = readList;

  //  cout << "in bufMgr ctor, " << numBufs << " of size " << bufferSize << " each " << endl;
  //  exit (0);

#ifdef LOG_HIT_RATE
  sprintf (logFileName,"/mnt/scsi_c/log/cfs/hitrate.log");
  cout << "Opening hit rate log" << logFileName << endl;
  fd = fopen (logFileName, "w");
  if (fd == NULL) {
    perror ("Failure opening hit rate log file");
    exit (0);
  }
#endif



#ifdef USE_SOLARIS
  pthread_mutexattr_t gattr;
  pthread_mutexattr_init(&gattr);
  pthread_mutexattr_settype(&gattr, PTHREAD_MUTEX_ERRORCHECK);
  pthread_mutex_init (&upl_mutex, &gattr);  
#else
  pthread_mutex_init (&upl_mutex, NULL);    
#endif

  //  pthread_mutex_init (&emptyUPL_mutex, NULL);    
  pthread_cond_init (&upl_cv, NULL);

  bufTable = new BufDesc[bufs];  // allocate array of buffer frame descriptors
  ASSERT (bufTable != NULL);

  for (int i = 0; i < bufs; i++) {
#ifdef USE_SOLARIS
    pthread_mutexattr_t mattr;
    pthread_mutexattr_init(&mattr);
    pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_ERRORCHECK);
    pthread_mutex_init (&(bufTable[i].mutex), &mattr);
#else
    pthread_mutex_init (&(bufTable[i].mutex), NULL);
#endif

    pthread_cond_init (&bufTable[i].pinCnt_cv, NULL);
    bufTable[i].frameNo = i;
  }

#ifdef USE_ALIGNED_MEMORY


  //  bufPoolMemory = new char [bufs * sizeof (Page) + BLOCK_SIZE];	 // allocate the actual buffer pool
  bufPoolMemory = new char [bufs * frameSize + BLOCK_SIZE];	 // allocate the actual buffer pool
  unsigned int ibuf;

  // align the buffer pool
  ibuf = (unsigned int)bufPoolMemory;
  if (ibuf % BLOCK_SIZE) {
    ibuf += (BLOCK_SIZE - (ibuf % BLOCK_SIZE));
  }
  
  bufPool = (Page *) ibuf;

  // for (int i = 0; i < bufs; i++) {
//     bufPool[i]->setDataPtr ((char*)(bufPool[i])); 
//   }


  //  int numBytes = bufs * sizeof (Page);
  ASSERT (frameSize == sizeof (Page));

  int numBytes = bufs * frameSize;
  int numKBytes = numBytes / 1024;
  int numMBytes = numKBytes / 1024;

  cout << "Aligned memory, " << bufs << " buffers of " << frameSize << " each " << endl;
  cout << "Buffer Pool is " << numMBytes << "M." << endl;

  ASSERT (numMBytes > 0);


#else
    bufPool = new Page[bufs];
#endif

  memset(bufPool, 0, bufs * sizeof(Page));  // zero out all buffer pool frames

  // wrong!
  int htsize = ((((int) ((double)bufs * 1.2))*2)/2)+1; // all to make sure an odd number


  cout << "htsize = " << htsize << endl;

  hashTable = new BufHashTbl (htsize); // allocate the buffer hash table

  //thread all pages in the pool together on the pinned list;
  // onto a two-way linked list with a sentinal node
  rearUPL = &bufTable[0]; // set rear pointer
  frontUPL = new BufDesc; // allocate a sentinal node

  // first set forward links
  for (int i=0; i < bufs-1; i++)
  {
	bufTable[i].next = &bufTable[i+1];   // forward link
  }
  bufTable[bufs-1].next = frontUPL;
  frontUPL->next = rearUPL;

  // then set reverse links
  for (int i=1; i < bufs; i++)
  {
	bufTable[i].prior = &bufTable[i-1];   // reverse link
  }
  bufTable[0].prior = frontUPL;
  frontUPL->prior = &bufTable[bufs-1];
  ASSERT (rearUPL->next != NULL);
  ASSERT (rearUPL->prior == frontUPL);
  ASSERT (frontUPL->next == rearUPL);

  // for debugging the request queue
  BufMgr::lowFrameAddress = &bufPool[0];
  BufMgr::highFrameAddress = &bufPool[bufs-1];

  returnStatus = OK;

}

//=============================================================================
BufMgr::~BufMgr() {

#ifdef LOG_HIT_RATE
  fclose (fd);
#endif

  cout << "Skipping sending dirty buffers to disk (out of " << numBufs << " buffers)" << endl;

//   // flush out all unwritten pages
//   for (int i = 0; i < numBufs; i++) 
//     {
//       cout << i << " " << endl;
//       BufDesc* tmpbuf = &bufTable[i];
//       if (tmpbuf->valid == true && tmpbuf->dirty == true) {

// #ifdef DEBUGBUF
// 	cout << "flushing page " << tmpbuf->pageNo
// 	     << " from frame " << i << endl;
// #endif

// 	cout << tmpbuf->pageNo << "(" << i << ") ";

// 	ASSERT (i >= 0 && i < numBufs);
// 	tmpbuf->file->writePage(tmpbuf->pageNo, &(bufPool[i]));
//       }
//     }
//   cout << endl;

  delete readList;

  delete [] bufTable;

#ifdef USE_ALIGNED_MEMORY
  delete [] bufPoolMemory;
#else
  delete [] bufPool;
#endif

  cout << "Freed buffer pool." << endl;

  delete hashTable;
  delete frontUPL;  // delete the sentinal node
}

//===============================================================================
const Status BufMgr::dumpHitRate () {
#ifdef LOG_HIT_RATE
  for (int i = 0; i < numBufs; i++) {
    BufDesc* tmpbuf = &bufTable[i];
    // only log hits during read phase
    if (tmpbuf->dirty == false && tmpbuf->valid == true) {
      int objCount = bufPool[tmpbuf->frameNo].getRecCnt();
      double hitRate = (double)tmpbuf->hits / (double)objCount;
      fprintf (fd, "%2.5f\t%d\t%d\n", hitRate, tmpbuf->hits, objCount);
    }
  }
#endif
    return OK;
}

//===============================================================================
const Status BufMgr::flush () {

  //   // flush out all unwritten pages

  Status status = OK;

  cout << "Flushing..." << endl;

  for (int i = 0; i < numBufs; i++) 
    {
      //      cout << i << " " << endl;
      BufDesc* tmpbuf = &bufTable[i];
      if (tmpbuf->valid == true) {
	if ((status = hashTable->remove(tmpbuf->file, tmpbuf->pageNo)) != OK) {
	  ASSERT (false);
	}

	if (tmpbuf->dirty == true) {
	  //	  cout << tmpbuf->pageNo << "(" << i << ") ";
	  ASSERT (i >= 0 && i < numBufs);
	  tmpbuf->file->writePage(tmpbuf->pageNo, &(bufPool[i]));
	  tmpbuf->dirty = false;
	}
	tmpbuf->valid = false;
	tmpbuf->hits = 0;
      }
    }
  cout << endl;

  return OK;
}

//=============================================================================
// return from allocBuf with our new frame locked
// enter here with upl_mutex locked
const Status BufMgr::allocBuf(int& frame) 
{
  Status status;
  BufDesc* victim;

#ifdef DEBUGBUF
  cout << pthread_self() << " in allocBuf" << endl;
#endif

  while (rearUPL == frontUPL) {
#ifdef DEBUGBUF
  cout << pthread_self() << " sleeping on empty UPL" << endl;
#endif
    pthread_cond_wait (&upl_cv, &upl_mutex);
#ifdef DEBUGBUF
  cout << pthread_self() << " woke on empty UPL" << endl;
#endif
  }

  // take a buffer from the rear of the unpinned list
  ASSERT (rearUPL != frontUPL);
  ASSERT (rearUPL != NULL);
  victim = rearUPL;
  ASSERT(victim->pinCnt == 0);

  // lock buffer
  int mutexCode = pthread_mutex_lock (&victim->mutex);
  ASSERT (mutexCode == 0);

  // remove entry from linked list
  rearUPL->next->prior = rearUPL->prior;
  rearUPL->prior->next = rearUPL->next;
  rearUPL = rearUPL->next;
  ASSERT (rearUPL->next != NULL);
  ASSERT (rearUPL->prior == frontUPL);
  ASSERT (frontUPL->next == rearUPL);
  frame = victim->frameNo;  // return frame

  if (victim->valid == false)          // page frame was not in use
  {
    victim->Clear();
    pthread_mutex_unlock (&upl_mutex);
    return OK;
  }
  else 
  {

#ifdef LOG_HIT_RATE
    // only log hits during read phase
    if (victim->dirty == false) {
      int objCount = bufPool[victim->frameNo].getRecCnt();
      double hitRate = (double)victim->hits / (double)objCount;
      fprintf (fd, "%2.5f\t%d\t%d\n", hitRate, victim->hits, objCount);
    }
#endif


    if ((status = hashTable->remove(victim->file, victim->pageNo)) != OK) {
      ASSERT (false);
      return status;
    }

    pthread_mutex_unlock (&upl_mutex);

      // else page frame was in use
      if (victim->dirty == true) 
      {    
	// page has unwritten data on it

#ifdef DEBUGBUF
	cout << "kicking out page " << victim->pageNo
             << " from frame " << victim->frameNo << endl;
#endif

	//	pthread_mutex_unlock (&upl_mutex);

	ASSERT (victim->frameNo >= 0 && victim->frameNo < numBufs);

	if ((status = victim->file->writePage(victim->pageNo, &(bufPool[victim->frameNo]))) != OK)
	  	return status;
	victim->dirty = false;

	//	pthread_mutex_lock (&upl_mutex);

      } 
      victim->Clear();
      return OK;
  }
}

//=============================================================================
// release an unused buffer frame - used in error conditions
const void BufMgr::releaseBuf(int frame) 
{

  ASSERT (false);

    bufTable[frame].next = rearUPL;
    bufTable[frame].prior = rearUPL->prior;
    rearUPL->prior->next = &bufTable[frame];
    rearUPL->prior = &bufTable[frame];
    rearUPL  = &bufTable[frame];
    ASSERT (rearUPL->next != NULL);
    ASSERT (rearUPL->prior == frontUPL);
    ASSERT (frontUPL->next == rearUPL);
}

//=============================================================================
// we leave here with the upl locked
const Status BufMgr::readPage(File* file, const int PageNo, Page*& page, const bool exclusive)
{
  int index;
  int mutexCode;
  Status status;

#ifdef DEBUGBUF
  cout << pthread_self() << " reading " << PageNo;
  if (exclusive)
    cout << " EX";
  else 
    cout << " SH";
  cout << endl;
#endif

  if (!file) {
    ASSERT (false);
    return BADFILE;
  }

  bool bothLocks = false;
  bool inHashTable = false;
  bool haveUPLlock = false;
  bool addedToReadList = false;

  while (!bothLocks) {
    inHashTable = false;

#ifdef DEBUGBUF
    cout << pthread_self() << " getting upl lock." << endl;
#endif

    if (!haveUPLlock)
      pthread_mutex_lock (&upl_mutex);

    if (!addedToReadList) {
      //      status = readList.add (file, PageNo);
      while ( (status = readList->add (file, PageNo)) == IN_READ_QUEUE) {
// #ifdef DEBUG_READLIST
// 	cout << pthread_self() << " returned from readList->add (read)" << endl;
// #endif
	pthread_mutex_unlock (&upl_mutex);
#ifdef DEBUG_READLIST
	cout << pthread_self() << " yielding due to readList collision (read)" << endl;
#endif
	sched_yield ();
#ifdef DEBUG_READLIST
	cout << pthread_self() << " awake again after readList collision (read)" << endl;

	//	sleep (1);
#endif
	pthread_mutex_lock (&upl_mutex);    
      }
      ASSERTSTATUS (status == OK);
      ASSERT (readList->verify (file, PageNo) == OK);
      addedToReadList = true;
    }

    if (hashTable->lookup(file, PageNo, index) == HASHNOTFOUND) {
      // we don't need both locks
      // we keep the upl lock on our way to allocBuf
#ifdef DEBUGBUF
      cout << pthread_self() << " did not find " << PageNo << " in hashtable" << endl;
#endif
      bothLocks = true;
    } else {
      // it's in the buffer pool in some state (on upl, shared, exclusive)
#ifdef DEBUGBUF
      cout << pthread_self() << " found " << PageNo << " in hashtable" << endl;
#endif
      inHashTable = true;

      if (pthread_mutex_trylock (&bufTable[index].mutex) == 0) {
	// we got the lock
	if (exclusive) {
#ifdef DEBUGBUF
	  cout << pthread_self() << " waiting for EX access to index " << index << " for pageNo " << PageNo << endl;
#endif
	  if (bufTable[index].pinCnt > 0) {
	    pthread_mutex_unlock (&upl_mutex);	    
	    haveUPLlock = false;
	    pthread_cond_wait (&bufTable[index].pinCnt_cv, &bufTable[index].mutex);
	    pthread_mutex_unlock (&bufTable[index].mutex);
	    // return to top of while loop and try to acquire locks in global order
	  } else {
	    bothLocks = true;  
	  }
	} else {
#ifdef DEBUGBUF
	  cout << pthread_self() << " got SH access to index " << index << " for pageNo " << PageNo << endl;
#endif
	  bothLocks = true;
	}
      } else {
	// we did not get the buffer lock
#ifdef DEBUGBUF
	  cout << pthread_self() << " sleeping on upl" << endl;
#endif
	pthread_cond_wait (&upl_cv, &upl_mutex);
	haveUPLlock = true;
	// again, return to top of loop
      }
    }
  }
  
  if (!inHashTable)
  {

    // page was not found in the hashtable
    if ((status = allocBuf(index)) != OK) {
      ASSERTSTATUS (false);
      return status;
    }

#ifdef DEBUG_READLIST
    cout << "Reading in pageNo " << PageNo << " from disk into index " << index << endl;
#endif

    //    pthread_mutex_unlock (&upl_mutex);

    ASSERT (readList->verify (file, PageNo) == OK);

    if ((status = file->readPage(PageNo, &(bufPool[index]))) != OK)
    {
      ASSERTSTATUS (false);
      releaseBuf(index); // release allocated frame 
         return status;
    }

    pthread_mutex_lock (&upl_mutex);

    if ((status = hashTable->insert(file, PageNo, index)) != OK)
    {
      ASSERTSTATUS (false);
      releaseBuf(index); // release allocated frame 
      return status;
    }
    bufTable[index].Set(file, PageNo);
  } 
  else 
  {
    // for pool with one buffer
    //    ASSERT (index == 0);
    // page was found in the buffer pool

    bufTable[index].hits++;

    //    ASSERT (bufTable[index].pinCnt == 0);

    if (bufTable[index].valid == false) {
      ASSERT (false);
      return BADBUFFER;
    }
    if (bufTable[index].pinCnt == 0)
    {
	// page is not pinned which means that it on the doubly linked 
	// list of unpinned pages.  remove it from the list
	// check to see if removing the rear node on the unpinned list
	// if so, fix up rearUPL appropriately before doing the remove
	if (rearUPL == &bufTable[index]) rearUPL = rearUPL->next;
	bufTable[index].prior->next = bufTable[index].next;
	bufTable[index].next->prior = bufTable[index].prior;
	bufTable[index].next = bufTable[index].prior = NULL;
  	ASSERT (rearUPL->next != NULL);
  	ASSERT (rearUPL->prior == frontUPL);
  	ASSERT (frontUPL->next == rearUPL);
    }
    bufTable[index].pinCnt++;

  }

  // return pointer to the page
  page = &(bufPool[index]);

  if (!exclusive) {
    mutexCode = pthread_mutex_unlock (&bufTable[index].mutex);
    ASSERT (mutexCode == 0);
  }

  status = readList->remove (file, PageNo);
  ASSERTSTATUS (status == OK);

  // unlock upl
  mutexCode = pthread_mutex_unlock (&upl_mutex);
  ASSERT (mutexCode == 0);
  

#ifdef DEBUGBUF
    cout << pthread_self() << " reading PageNo " << PageNo
         << " from Buffer " << index  
         << " pinCnt is " << bufTable[index].pinCnt << endl;
#endif DEBUGBUF

  return OK;
}

//=============================================================================
// we should enter here with the upl locked
const Status BufMgr::unPinPage(File* file, const int PageNo, 
			       const bool dirty, const bool loveIt, const bool exclusive)
{
  int index;
  int mutexCode;
  //  bool uplWasEmpty = false;

#ifdef DEBUGBUF
  cout << pthread_self() << " unPinning " << PageNo;
  if (exclusive)
    cout << " EX";
  else 
    cout << " SH";
  cout << endl;
#endif


  // lock upl
  mutexCode = pthread_mutex_lock (&upl_mutex);
  ASSERT (mutexCode == 0);

  // see if page is in the hash table
  if (hashTable->lookup(file, PageNo, index) == HASHNOTFOUND) {
    ASSERT (false);
    return HASHNOTFOUND;
  }

    // for pool with one buffer
  //    ASSERT (index == 0);

  if (!exclusive) {
    mutexCode = pthread_mutex_lock (&bufTable[index].mutex);
    ASSERT (mutexCode == 0);
  }
  // if exclusive, we already have a lock on this buffer
  // if shared, nobody should have an exclusive lock on this buffer, 
  // so we could only wait if somebody were getting a shared lock
  // or releasing a shared lock on this buffer

  if (bufTable[index].valid == false) {
    mutexCode = pthread_mutex_unlock (&upl_mutex);
    ASSERT (mutexCode == 0);
    ASSERT (false);
    return BADBUFFER;
  }
  if (bufTable[index].pinCnt == 0) {
    mutexCode = pthread_mutex_unlock (&upl_mutex);
    ASSERT (mutexCode == 0);
    ASSERT (false);
    return PAGENOTPINNED;
  }

  if (dirty == true) bufTable[index].dirty = dirty;
  bufTable[index].pinCnt--;
  if (bufTable[index].pinCnt == 0)
  {
    // upl is currently empty and there are waiters in allocBuf
    // if (rearUPL == frontUPL) {
//       uplWasEmpty = true;
//     }

      if (loveIt) 
      {
	// page is LOVED, add it to the front of the unpinned list
	// simulates LRU
	bufTable[index].next = frontUPL;
	bufTable[index].prior = frontUPL->prior;
	frontUPL->prior->next = &bufTable[index];
	frontUPL->prior = &bufTable[index];
	// in the case that the list was empty need to reset rear pointer
	if (frontUPL == rearUPL) rearUPL = &bufTable[index];
  	ASSERT (rearUPL->next != NULL);
  	ASSERT (rearUPL->prior == frontUPL);
  	ASSERT (frontUPL->next == rearUPL);
      }
      else
      {
	// page is NOT LOVED, so add it to the rear of the unpinned list
	// so it will be allocated first - simulates MRU
	// add to rear of list
	bufTable[index].next = rearUPL;
	bufTable[index].prior = rearUPL->prior;
	rearUPL->prior->next = &bufTable[index];
	rearUPL->prior = &bufTable[index];
	rearUPL  = &bufTable[index];
  	ASSERT (rearUPL->next != NULL);
  	ASSERT (rearUPL->prior == frontUPL);
  	ASSERT (frontUPL->next == rearUPL);
      }
  }

  pthread_cond_signal (&bufTable[index].pinCnt_cv);
  mutexCode = pthread_mutex_unlock (&bufTable[index].mutex);
  ASSERT (mutexCode == 0);


//   if (uplWasEmpty) {
#ifdef DEBUGBUF
    cout << pthread_self() << " signalling in unPinPage" << endl;
#endif
    pthread_cond_signal (&upl_cv);
//   }

  mutexCode = pthread_mutex_unlock (&upl_mutex);    
  ASSERT (mutexCode == 0);

#ifdef DEBUGBUF
  cout << pthread_self();
  if (exclusive)
    cout << " EX ";
  else
    cout << " SH ";
  cout << " page unlock (unPinPage) on page " << bufTable[index].pageNo << " in frame " << index << ", pinCnt is " << bufTable[index].pinCnt << endl;
#endif

  return OK;
}

//=============================================================================
// leave here with upl and buffer locked
const Status BufMgr::allocPage(File* file, int& pageNo, Page*& page) {

  Status status;
  int mutexCode;

#ifdef DEBUGBUF
  cout << pthread_self() << " in allocPage ";
#endif

  pthread_mutex_lock (&upl_mutex);    

  if ((status = file->allocatePage(pageNo)) != OK) {
    cout << "Problem in BufMgr::allocPage" << endl;
    //    ASSERTSTATUS (false);
    return status;
  }

//    while (readList->add (file, pageNo) == IN_READ_QUEUE) {
//      pthread_mutex_unlock (&upl_mutex);
// #ifdef DEBUG_READLIST
//      cout << pthread_self() << " yielding due to readList collision (alloc)" << endl;
// #endif
//      sched_yield ();
//      pthread_mutex_lock (&upl_mutex);    
//    }

  int index;

  status = allocBuf(index); // allocate a buffer
  if (status != OK || index >= numBufs) {
    ASSERTSTATUS (false);
    return status;
  }

  pthread_mutex_lock (&upl_mutex);

  if ((status = hashTable->insert(file, pageNo, index)) != OK) 
  {
    ASSERTSTATUS (false);
      releaseBuf(index);
      return status;
  }

  bufTable[index].Set(file, pageNo);

//  bufTable[index].dirty = true;      // Is this needed ?
  page = &(bufPool[index]);

  //  readList->remove (file, pageNo);

 // unlock upl
  mutexCode = pthread_mutex_unlock (&upl_mutex);
  ASSERT (mutexCode == 0);

#ifdef DEBUGBUF
  cout << pthread_self() << " allocating pageNo " << pageNo   
       << " to Buffer " << index  
       << "\tpinCnt is " << bufTable[index].pinCnt << "\n";
#endif DEBUGBUF

  return OK;
}

//=============================================================================
int BufMgr::getFrameCount() {
  return numBufs;
}


// buffer pool hash table implementation

int BufHashTbl::hash(const File* file, const int pageNo)
{
  int tmp, value;
  tmp = (int)file;  // cast of pointer to the file object to an integer
  value = (tmp + pageNo) % HTSIZE;
  return value;
}


BufHashTbl::BufHashTbl(int htSize)
{

  pthread_mutex_init (&mutex, NULL);
  HTSIZE = htSize;
  // allocate an array of pointers to hashBuckets
  ht = new BufHashBucket* [htSize];
  for(int i=0; i < HTSIZE; i++)
    ht[i] = NULL;
}


BufHashTbl::~BufHashTbl()
{
  for(int i = 0; i < HTSIZE; i++) {
    BufHashBucket* tmpBuf = ht[i];
    while (ht[i]) {
      tmpBuf = ht[i];
      ht[i] = ht[i]->next;
      delete tmpBuf;
    }
  }
  delete [] ht;
}


//---------------------------------------------------------------
// insert entry into hash table mapping (file,pageNo) to frameNo;
// returns OK if OK, BUFHASHTBLERROR if an error occurred
//---------------------------------------------------------------

Status BufHashTbl::insert(const File* file, const int pageNo, const int frameNo) {

  int index = hash(file, pageNo);

  //  pthread_mutex_lock (&mutex);

#ifdef DEBUG_BUFHASHTBL
  //  pthread_t pt = pthread_self();
  cout << pthread_self() << " inserting " << file->name << " pageNo " << pageNo << " into frameNo " << frameNo << " in BufHashTbl" << endl;
#endif

  BufHashBucket* tmpBuc = ht[index];
  while (tmpBuc) {
    if (tmpBuc->file == file && tmpBuc->pageNo == pageNo) {
      //      pthread_mutex_unlock (&mutex);
      return BUFHASHTBLERROR;
    }
    tmpBuc = tmpBuc->next;
  }

  tmpBuc = new BufHashBucket;
  if (!tmpBuc) {
    //    pthread_mutex_unlock (&mutex);
    return BUFHASHTBLERROR;
  }
  tmpBuc->file = (File*) file;
  tmpBuc->pageNo = pageNo;
  tmpBuc->frameNo = frameNo;
  tmpBuc->next = ht[index];
  ht[index] = tmpBuc;


  //  pthread_mutex_unlock (&mutex);
  return OK;
}


//-------------------------------------------------------------------	     
// Check if (file,pageNo) is currently in the buffer pool (ie. in
// the hash table).  If so, return corresponding frameNo. else return 
// HASHNOTFOUND
//-------------------------------------------------------------------

const Status BufMgr::markDirty (const File* file, const int pageNo) {
  Status status;
  int frameNo;
  status = hashTable->lookup (file, pageNo, frameNo);
  if (status != OK) return status;
#ifdef DEBUGBUF
  cout << "Marking frame " << frameNo << " dirty." << endl;
#endif
  bufTable[frameNo].dirty = true;
  return OK;
}

Status BufHashTbl::lookup(const File* file, const int pageNo, int& frameNo) 
  {
  int index = hash(file, pageNo);

  //  pthread_mutex_lock (&mutex);

#ifdef DEBUG_BUFHASHTBL
  cout << pthread_self() << " looking up " << file->name << " pageNo " << pageNo << " in BufHashTbl" << endl;
#endif

  BufHashBucket* tmpBuc = ht[index];
  while (tmpBuc) {
    if (tmpBuc->file == file && tmpBuc->pageNo == pageNo)
    {
      frameNo = tmpBuc->frameNo; // return frameNo by reference
      //      pthread_mutex_unlock (&mutex);
      return OK;
    }
    tmpBuc = tmpBuc->next;
  }
  //  pthread_mutex_unlock (&mutex);
  return HASHNOTFOUND;
}


//-------------------------------------------------------------------
// delete entry (file,pageNo) from hash table. REturn OK if page was
// found.  Else return BUFHASHTBLERROR
//-------------------------------------------------------------------

Status BufHashTbl::remove(const File* file, const int pageNo) {

  int index = hash(file, pageNo);

  //  pthread_mutex_lock (&mutex);
#ifdef DEBUG_BUFHASHTBL
  cout << pthread_self() << " removing " << file->name << " pageNo " << pageNo << " in BufHashTbl" << endl;
#endif

  BufHashBucket* tmpBuc = ht[index];
  BufHashBucket* prevBuc = ht[index];

  while (tmpBuc) {
    if (tmpBuc->file == file && tmpBuc->pageNo == pageNo) {
      if (tmpBuc == ht[index]) 
	ht[index] = tmpBuc->next;
      else
	prevBuc->next = tmpBuc->next;
      delete tmpBuc;
      //      pthread_mutex_unlock (&mutex);
      return OK;
    } else {
      prevBuc = tmpBuc;
      tmpBuc = tmpBuc->next;
    }
  }

  //  pthread_mutex_unlock (&mutex);
  return BUFHASHTBLERROR;

}

ostream& operator << (ostream& os, BufDesc& bufDesc) {
     os << bufDesc.frameNo << "\t" << bufDesc.file << "\t";
     os << bufDesc.pageNo << "\t" << bufDesc.pinCnt << "\t";
     if (pthread_mutex_trylock(&bufDesc.mutex) == 0) {
       os << "unlock\t";
       pthread_mutex_unlock (&bufDesc.mutex);
     } else {
       os << "locked\t";
     }

     os <<  bufDesc.dirty << "\t" << bufDesc.valid << endl;
     return os;
}

int BufMgr::getPinnedPageCount () {

  int mutexCode = pthread_mutex_lock (&upl_mutex);
  ASSERT (mutexCode == 0);

  int cntAct = 0;
   BufDesc  *p;
   p = &(bufTable)[0];

  for (int i=0; i< numBufs; i++,p++)  {
    if (p->getPinCnt() > 0) cntAct++;
  }

  mutexCode = pthread_mutex_unlock (&upl_mutex);
  ASSERT (mutexCode == 0);

  return cntAct;
}

ostream& operator << (ostream& os, BufMgr& bufMgr) {

   os << "Buffer Manager:" << endl;

  int mutexCode = pthread_mutex_lock (&(bufMgr.upl_mutex));
  ASSERT (mutexCode == 0);

   int i; 
   BufDesc  *p;
   p = &(bufMgr.bufTable)[0];
   int cntAct = 0;



   os <<"frontUPL = " << bufMgr.frontUPL <<" rearUPL = " << bufMgr.rearUPL << "\n";
    os <<"frameNo\tfile\tpageNo\tpinCnt\tlockCnt\tdirty\tvalid\n";
    for (i=0; i<bufMgr.numBufs; i++,p++)
   {
     os << *p;

     if (p->getPinCnt() > 0) cntAct++;
   }
   // now count on the number of pages on the free list going forward
   p = bufMgr.rearUPL;
   int cntFwd = 0;
   while (p != bufMgr.frontUPL)
   {
	cntFwd++;
	p = p->getNext();
	if (p->getPinCnt() > 0) 
	 printf("ERROR. frame %d is on unpinned list but has a pinCnt of %d\n",
		p->getFrameNo(), p->getPinCnt());
   }
   // now count on the number of pages on the free list going backward
   p = bufMgr.frontUPL->getPrior();
   int cntBwd = 0;
   while (p != bufMgr.frontUPL) // stop when it wraps
   {
	cntBwd++;
	p = p->getPrior();
   }

   os << "# of pages actually pinned = " << cntAct << ", forward=%d, backward= " << cntBwd << endl;

   ASSERT (bufMgr.rearUPL->getNext() != NULL);
   ASSERT (bufMgr.rearUPL->getPrior() == bufMgr.frontUPL);
   ASSERT (bufMgr.frontUPL->getNext() == bufMgr.rearUPL);

  mutexCode = pthread_mutex_unlock (&(bufMgr.upl_mutex));
  ASSERT (mutexCode == 0);
  return os;
}

//=============================================================================
int BufDesc::getPinCnt () {
  return pinCnt;
}

int BufDesc::getFrameNo () {
  return frameNo;
}

BufDesc* BufDesc::getNext () {
  return next;
}

BufDesc* BufDesc::getPrior () {
  return prior;
}

//=============================================================================
ReadList::ReadList (int threadCount) : length (threadCount) {
  // we can never have more readers than threads
  reader = new Reader [length + 1];
  for (int i = 0; i < length + 1; i++) {
    reader[i].file = NULL;
    reader[i].pageNo = -1;
  }
}

ReadList::~ReadList () {
  delete [] reader;
}

//=============================================================================
Status ReadList::add (File *file, int pageNo) {

  int match = 0;
  int numFound = 0;
  for (int index = 0; index < length; index++) {
    if (reader[index].file != NULL)
      numFound++;
    if (reader[index].file == file && reader[index].pageNo == pageNo) {
      match++;
    }
  }
  ASSERT (match == 0 || match == 1);

  if (match == 1)
    return IN_READ_QUEUE;

  reader[numFound].file = file;
  reader[numFound].pageNo = pageNo;

  //  cout << *this;

  return OK;
  

}

//=============================================================================
Status ReadList::remove (const File *file, const int pageNo) {

  int match = 0;
  
  for (int index = 0; index < length; index++) {
    if (reader[index].file == file && reader[index].pageNo == pageNo) {
      match++;
    }
    // shift back
    if (match == 1) {
      reader[index].file = reader[index + 1].file;
      reader[index].pageNo = reader[index + 1].pageNo;
    }
  }

  ASSERT (match == 1);

  // cout << *this;

  return OK;
}

Status ReadList::verify (const File *file, const int pageNo) {
  static int timesThrough;
  static int maxLength;

  const int debugPeriod = 5000;

  int match = 0;
  int numFound = 0;

  if (++timesThrough % debugPeriod == 0) {
    cout << "read list: ";
  }

  for (int index = 0; index < length; index++) {
    if ( (timesThrough % debugPeriod == 0) && (reader[index].file != NULL)) {
      cout << "(" << (reader[index].file)->id << "," << reader[index].pageNo << ") ";
    }

    if (reader[index].file == file && reader[index].pageNo == pageNo) {
      match++;
    }
    if (reader[index].file != NULL) numFound++;
  }
  ASSERT (match == 1);

  if (numFound > maxLength) {
    maxLength = numFound;
    cout << "New readList length: " << maxLength << endl;
  }

  if (timesThrough % debugPeriod == 0) {
    cout << "\nIn ReadList::verify, " << timesThrough << " times.  Num found = " << numFound << ", length = " << length << endl;
  }

  return OK;
}

ostream& operator << (ostream& os, const Reader &reader) {
  os << "(" << *(reader.file) << "," << reader.pageNo << ")";
  return os;
}

ostream& operator << (ostream& os, const ReadList &readList) {

  os << "Reader: ";
  for (int index = 0; index < readList.length + 1; index++) {
    if (readList.reader[index].file == NULL)
      cout << "(" << index << ",nil," << readList.reader[index].pageNo << ") ";
    else
      cout << "(" << index << "," << (readList.reader[index].file)->id << "," << readList.reader[index].pageNo << ") ";
  }

  os << "\n";
  return os;
}

