/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef BUFFER_MANAGER_H
#define BUFFER_MANAGER_H

//#include <set>
#include <iostream>
#include "file.h"

// define if debug output wanted

// declarations for buffer pool hash table
struct BufHashBucket
{
	File*	file;    // pointer a file object (more on this below)
	int	pageNo;  // page number within a file
	int	frameNo; // frame number of page in the buffer pool
	BufHashBucket* 	next;	 // next node in the hash table
};


// hash table to keep track of pages in the buffer pool
class BufHashTbl
{
private:
    int HTSIZE;
    BufHashBucket**  ht; // actual hash table
    int	 hash(const File* file, const int pageNo); // returns value between 0 and HTSIZE-1
    pthread_mutex_t mutex;

public:
    BufHashTbl(const int htSize);  // constructor
    ~BufHashTbl(); // destructor
	
    // insert entry into hash table mapping (file,pageNo) to frameNo;
    // returns 0 if OK, HASHTBLERROR if an error occurred
  Status insert(const File* file, const int pageNo, const int frameNo);

    // Check if (file,pageNo) is currently in the buffer pool (ie. in
    // the hash table).  If so, return corresponding frameNo. else return 
    // HASHNOTFOUND
  Status lookup(const File* file, const int pageNo, int & frameNo);

    // delete entry (file,pageNo) from hash table. REturn OK if page was
    // found.  Else return HASHTBLERROR
  Status remove(const File* file, const int pageNo);  
};


class BufMgr;  //forward declaration of BufMgr class 

// class for maintaining information about buffer pool frames
class BufDesc {
    friend class BufMgr;
    friend ostream& operator << (ostream& os, BufDesc& bufDesc);

 public:
    BufDesc* getNext ();
    BufDesc* getPrior ();
    int getPinCnt ();
    int getFrameNo ();
  private:
  File* file;   // pointer to file object
  int   pageNo; // page within file
  int	frameNo;  // frame # of frame
  int   pinCnt; // number of times this page has been pinned

  int hits;

  //  pthread_mutexattr_t attr;
  pthread_mutex_t mutex;
  pthread_cond_t pinCnt_cv;

  bool 	dirty;	  // true if dirty;  false otherwise
  bool 	valid;   // true if page is valid
  BufDesc* next;  // next frame in doubly linked list of unpinned frames
  BufDesc* prior;  // prior frame in doubly linked list of unpinned frames

  void Clear() {  // initialize buffer frame for a new user
    hits = 1;
    	pinCnt = 0;
	file = NULL;
	pageNo = -1;
    	dirty = false;
	valid = false;
	next = prior = NULL;
  };

  void Set(File* filePtr, int pageNum) { 

    hits = 1;
      file = filePtr;
      pageNo = pageNum;
      pinCnt = 1;
      dirty = false;
      valid = true;
      next = prior = NULL;
  }

  BufDesc() {
      Clear();
  }

};

struct Reader {
  friend ostream& operator << (ostream& os, const Reader& reader);
  File *file;
  int pageNo;
};

// used to restrict readPage to only one file/pageNo at a time
// this is so that multiple buffers don't get allocated to the same file/pageNo
class ReadList {
  friend ostream& operator << (ostream& os, const ReadList& readList);
 private:
  //  Reader *reader;
  //  int length;
  //  set<Reader> reader;

  // array of readers
  Reader *reader;
  const int length;

 public:
  ReadList (int length);
  ~ReadList ();
  Status add (File *file, int pageNo);
  Status remove (const File *file, const int pageNo);
  Status verify (const File *file, const int pageNo);
};

class BufMgr 
{
  friend ostream& operator << (ostream& os, BufMgr& bufMgr);
private:

  BufHashTbl*    hashTable;  // hash table mapping (File, page) to frame
  BufDesc*	 bufTable;  // vector of status info, 1 per page
  BufDesc*	 rearUPL;    // pointer to the rear of the list of unpinned frames
  BufDesc*	 frontUPL;    // pointer to the front of the list of unpinned frames

  const Status allocBuf(int & frame);   // allocate a free frame.  
  const void releaseBuf(int frame); // return unused frame to end of list

  ReadList *readList;

  static ReadList *staticReadList;

  FILE* fd;
  char logFileName [100];

public:
  const int   	 numBufs;    // Number of pages in buffer pool
  char*          bufPoolMemory;
  Page*	         bufPool;   // actual buffer pool

  BufMgr (const int bufs, const int bufferSize, Status& returnStatus);
    //  BufMgr(const int bufs);
  ~BufMgr();

  int getPinnedPageCount ();

  const Status readPage(File* file, const int PageNo, Page*& page, const bool exclusive);
  const Status unPinPage(File* file, const int PageNo, const bool dirty, const bool loveIt, const bool exclusive);
  const Status allocPage(File* file, int& PageNo, Page*& page); // allocates a new, empty page 

  const Status flush ();
  const Status dumpHitRate ();

  int getFrameCount();

  // marks whatever page this file/pageNo is in dirty
  const Status markDirty (const File* file, const int pageNo);

  pthread_mutex_t upl_mutex;
  //  pthread_mutex_t emptyUPL_mutex;
  //  pthread_cond_t emptyUPL_cv;
  pthread_cond_t upl_cv;

  // statistics
  long allocBufCount;
  long allocBufSleep;
  long readPageCount;
  long readPageSleep;

 static Page *lowFrameAddress;    // used to check that data pointer is within the frame buffer's space
 static Page *highFrameAddress;
 static void printReadList ();
};

#endif
