/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "cfs.h"
#include <stdio.h>

int CFS::framesPerGroup;
int CFS::blocksPerFrame;
int CFS::frameCount;
int CFS::groupCount;
//int CFS::sectorCount;

void * runDisk (void * arg);

//=============================================================================
// The CFS class essentially converts and caches URLs to md5s, 
// and multiplexes requests out to the Groups, where the real work is done.

CFS::CFS (const char *const myRawDiskDevice, int myGroupCount, int myFrameCount, Status& returnStatus) : rawDiskDevice (myRawDiskDevice), quickDigestListLength(10) {

  CFS::groupCount = myGroupCount;
  CFS::frameCount = myFrameCount;
  CFS::blocksPerFrame = frameSize / BLOCK_SIZE;

  //  int availableDiskFrameCount = CFS::sectorCount / CFS::blocksPerFrame;
//   cout << sectorCount << " sectorCount " << endl;
//   cout << blocksPerFrame << " blocksPerFrame " << endl;
//   cout << availableDiskFrameCount << " availableDiskFrameCount" << endl;

  
 CFS::framesPerGroup = (diskSectorCount / CFS::blocksPerFrame)  / CFS::groupCount;


 cout << framesPerGroup << "  disk frames per group " << endl;

  // int diskFrameCount = CFS::framesPerGroup * CFS::groupCount;
//   cout << diskFrameCount << " diskFrameCount " << endl;

  // number of buffers
  bufMgr = new BufMgr (frameCount, frameSize, returnStatus);
  if (returnStatus != OK) return;

  // pop the disk into its own thread
  pthread_t diskThread;

  disk = new Disk (rawDiskDevice, returnStatus);
  if (returnStatus != OK) return;
  pthread_create (&diskThread, NULL, &runDisk, (void *) disk);

  // turn group into 
  group = new (Group *) [groupCount];

  int currentOffset = 0;

  // number of groups
  for (int i = 0; i < groupCount; i++) {
    group[i] = new Group (i, bufMgr, disk->getRequestQueue(), currentOffset, framesPerGroup, returnStatus);
    if (returnStatus != OK) return;
    group[i]->setLoveIt (LOVE);
    currentOffset += framesPerGroup;
  }

#ifdef QUICK_CFS_LOOKUP
  fileTable = new (FileTable *)[groupCount];
  for (int i = 0; i < groupCount; i++) {
    fileTable[i] = group[i]->getFileTable ();
  }
#endif


#ifdef QUICK_MD5
  pthread_mutex_init (&quickMutex, NULL);
  quickDigestListFront = NULL;
#endif

  returnStatus = OK;

}

//=============================================================================
void * runDisk (void * arg) {
  ((Disk*)arg)->run ();
  return NULL;
}

//=============================================================================
CFS::~CFS () {

  cout << "Shutting down Cache File System..." << endl;

#ifdef QUICK_CFS_LOOKUP
  delete [] fileTable;
#endif  

#ifdef QUICK_MD5
  QuickDigest *current = quickDigestListFront;
  while (current != NULL) {
    QuickDigest *deleteMe = current;
    current = current->next;
    delete deleteMe;
  }
#endif

  delete bufMgr;

  for (int i = 0; i < groupCount; i++) {
    delete group[i];
  }
  delete [] group;
  delete disk;

}

#ifdef QUICK_CFS_LOOKUP

//=============================================================================

const Status CFS::lookup (const char *const url, int & groupIndex) {

  Status status = NULL_STATUS;
  md5_byte_t *digest = new md5_byte_t [16];
  genMD5 (url, digest);

  int chainIndex = FileTable::hash (digest);

  //   // traverse groups' FileTables for match
  for (groupIndex = 0; (groupIndex < groupCount && status != OK); groupIndex++) {
    status = fileTable[groupIndex]->lookup (digest, chainIndex);
    // we found it
    if (status == OK) {
      delete [] digest;
      return OK;
    }
  }

  groupIndex = NULL_GROUP;
  delete [] digest;
  return FILE_NOT_FOUND;

}

#else

//=============================================================================

const Status CFS::lookup (const char *const url, int & groupIndex) {

  Status status = NULL_STATUS;
  md5_byte_t *digest = new md5_byte_t [16];
  genMD5 (url, digest);

  //   // traverse groups' FileTables for match
  for (groupIndex = 0; (groupIndex < groupCount && status != OK); groupIndex++) {
    //     status = group[groupIndex]->lookup (digest, rid);
    status = group[groupIndex]->lookupRecord (digest);
    // we found it
    if (status == OK) {
      //     // insert it into our FIFO queue of md5s (i.e. a cache of 'em)
      delete [] digest;
      return OK;
    }
  }

  groupIndex = NULL_GROUP;
  delete [] digest;
  return FILE_NOT_FOUND;

}

#endif

//=============================================================================
// leaves the page dirty, unpinned, and group->loved

const Status CFS::create (const char *const url, const int groupIndex, char * data, int length) {

  Status status;
  // change url -> digest
  md5_byte_t *digest = new md5_byte_t [16];
  genMD5 (url, digest);



  // turn the data/size into a Rec
  Record rec;
  rec.data = data;
  rec.length = length;

  status = group[groupIndex]->insertRecord (digest, rec);

  delete [] digest;

  return status;
}

//=============================================================================

const Status CFS::read (const char *const url, const int groupIndex, char *&data, int &length) {

  // leaves the page pinned in the buffer pool
  Status status;

  // change url -> digest
  md5_byte_t *digest = new md5_byte_t [16];
  genMD5 (url, digest);

  Record rec;
  status = group[groupIndex]->readRecord (digest, rec);

  if (status != OK) {
    delete [] digest;
    return status;
  }

  // change the Rec into data/size
  data = (char *) rec.data;
  length = rec.length;

  delete [] digest;
  //  ASSERT (bufMgr->getPinnedPageCount() == 1);
  return OK;
}

//=============================================================================
// unpins the page this URL is on (without modifying dirty bit)
// this is only called after a CFS::Read
//

const Status CFS::release (const char *const url, const int groupIndex) {

  Status status;
  // change url -> digest
  md5_byte_t *digest = new md5_byte_t [16];
  genMD5 (url, digest);

  status = group[groupIndex]->releaseRecord (digest);
  delete [] digest;

  //  ASSERT (bufMgr->getPinnedPageCount() == 0);
  return status;
}

//=============================================================================

const Status CFS::remove (const char *const url, const int groupIndex) {

  Status status;
  // change url -> digest
  md5_byte_t *digest = new md5_byte_t [16];
  genMD5 (url, digest);

  status = group[groupIndex]->removeRecord (digest);
  delete [] digest;

  //  ASSERT (bufMgr->getPinnedPageCount() == 0);
  return OK;
}

//=============================================================================
const Status CFS::love (const int groupIndex) {
  return group[groupIndex]->setLoveIt (LOVE);
}


//=============================================================================
const Status CFS::hate (const int groupIndex) {
  return group[groupIndex]->setLoveIt (HATE);
}

//=============================================================================
void CFS::genMD5 (const char *url, md5_byte_t *digest) {
  //void CFS::genMD5 (const char *url, char *digest) {

#ifdef QUICK_MD5
  pthread_mutex_lock (&quickMutex);
  int currentLength = 0;
  QuickDigest *current = quickDigestListFront;
  QuickDigest *last = NULL;
  QuickDigest *penultimate = NULL;
  //  string urlString = url;
  while (current != NULL) {
    currentLength++;
    //    if (urlString == current->url) {
    if (strcmp (url, current->url) == 0) {
      // cout << "quick match" << endl;
      //      memcpy (digest, current->digest, DIGEST_LENGTH);
      memcpy (digest, current->digest, sizeof (md5_byte_t [16]));
      pthread_mutex_unlock (&quickMutex);
      return;
    }
    penultimate = last;
    last = current;
    current = current->next;
  }
  //  pthread_mutex_unlock (&quickMutex);
#endif

  md5_state_t state;
  //        md5_byte_t digest[16];

  md5_init(&state);
  md5_append(&state, (const md5_byte_t *)url, strlen(url));
  md5_finish(&state, digest);

//   printf("MD5 (\"%s\") =\n", url);
  //  for (int di = 0; di < 16; ++di)
  //     printf("%02x", digest[di]);
  //   printf("\n");

#ifdef QUICK_MD5
  //  pthread_mutex_lock (&quickMutex);
  QuickDigest *newQuick = new QuickDigest (url, digest);
  newQuick->next = quickDigestListFront;
  quickDigestListFront = newQuick;

  if (currentLength > quickDigestListLength) {
    // cout << "remove last in quick list" << endl;
    if (last != NULL && penultimate != NULL) {
      penultimate->next = NULL;
      delete [] last->url;
      delete last;
    }
  }
  pthread_mutex_unlock (&quickMutex);
#endif

  return;
}

//=============================================================================
QuickDigest::QuickDigest (const char *urlString, md5_byte_t *digest) {
  //  this->url = url;
  url = new char [strlen (urlString)];
  ASSERT (url != NULL);
  memcpy (url, urlString, strlen (urlString));
  memcpy (this->digest, digest, sizeof (md5_byte_t [16]));
  next = NULL;
}

QuickDigest::~QuickDigest () {
}

//=============================================================================
void CFS::printMD5 (md5_byte_t *digest) {
  int di;
  cout << "size = " << sizeof (md5_byte_t *) << endl;

  for (di = 0; di < 16; ++di)
    printf("%02x", digest[di]);
  printf("\n");
}

//=============================================================================

ostream& operator << (ostream& os, CFS* &cfs) 
{
  //  os << "File system:";
  //  os << "\n\treads: " << cfs->filesRead;
  //  os << "\n\twrites: " << cfs->filesWritten;
  BufMgr *bufMgr = cfs->bufMgr;
  os << "pincnt = " << bufMgr->getPinnedPageCount() << "\n";
  //  os << "\n" << *(cfs->bufMgr);
  Disk *disk = cfs->disk;
  os << *disk;
  return os;
}

//=============================================================================
BufMgr* CFS::getBufMgr () {
  return bufMgr;
}

//=============================================================================
const Status CFS::flushBufferPool () {
  return bufMgr->flush();
}

//=============================================================================
const Status CFS::dumpHitRate () {
  return bufMgr->dumpHitRate();
}

