/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "disk.h"

int Disk::diskFrameCount;

Disk::Disk (const char *fileName, Status& returnStatus) {
  head = 0;
  maxRequestQueue = 0;

  Disk::diskFrameCount = diskSectorCount / CFS::blocksPerFrame;

//   cout << "Disk can hold " << frames << " frames of " << frameSize << " bytes each." << endl;
  //  int frameSize = frameSize;

  int totalBytes = diskSectorCount * BLOCK_SIZE;
  cout << totalBytes << " total bytes available in disk:" << fileName << endl;
  cout << diskFrameCount << " disk frames of " << frameSize << " each" << endl;

  int diskBytes = diskFrameCount * frameSize;

  ASSERT (totalBytes == diskBytes);

  requestQueue = new RequestQueue (&head);

  unlink (fileName);

  // create new file
  if( (fd = open(fileName, O_RDWR)) < 0) {  
    if((fd = open(fileName, O_RDWR | O_CREAT | O_TRUNC, S_IWUSR | S_IRUSR )) < 0){
      cout << "Fatal error opening disk!\n";
      returnStatus = DISK_ERROR;
      return;
    }
  }

 if (lseek(fd, BLOCK_SIZE, SEEK_SET) < 0) {
   perror ("Disk::Disk: ");
   returnStatus = DISK_ERROR;   
   return;
 }

  quitting = false;
  returnStatus = OK;
}

Disk::~Disk () {
  close (fd);
  delete requestQueue;
  quitting = true;
}

//=============================================================================
void Disk::run () {
#ifdef DEBUG_DISK
    cout << "Running disk" << endl;
#endif

  Status status;

  // main loop for disk thread
  while (1) {
  pthread_mutex_lock (&requestQueue->mutex);
  while (requestQueue->length == 0) {
    pthread_cond_wait (&requestQueue->cv, &requestQueue->mutex);
  }

#ifdef DEBUG_DISK
    cout << "disk got a request, queue length " << requestQueue->length << endl;
#endif

  if (maxRequestQueue < requestQueue->length)
    maxRequestQueue = requestQueue->length;

  Request *request;
  if ( (status = requestQueue->dequeue (request)) != OK) {
    cout << status << endl;
    ASSERT (false);
  }

#ifdef DEBUG_DISK
    cout << "disk dequeued ok" << endl;
    cout << "request = " << request << endl;
#endif

  pthread_mutex_unlock (&requestQueue->mutex);

  switch (request->action) {
  case WRITE:
    request->status = doWrite (request);
    break;
  case READ:
    request->status = doRead (request);
    break;
  default:
    cout << "Unknown action on request queue" << endl;
    break;
  }

#ifdef DEBUG_DISK
    cout << "Finished request, waking: " << request << endl;
#endif

  pthread_mutex_lock(&request->mutex);
  request->complete = true;
  pthread_cond_signal (&request->cv);
  pthread_mutex_unlock(&request->mutex);

#ifdef SLOW_DISK
  sleep (SLOW_DISK);
#endif

#ifdef DEBUG_DISK
  cout << "woke: " << request << endl;
#endif

  if (quitting)
    pthread_exit (0);
 
  }

}

//=============================================================================
Status Disk::doWrite (Request *request) {
  // Status status;

#ifdef DEBUG_DISK
  cout << "Starting write request " << request << endl;
#endif

  // set the disk head to the right location on disk

  if (head != request->frameNumber) {
    // keep totalSeekDistance in blocks, not bytes
    int byteOffset = request->frameNumber * CFS::blocksPerFrame * BLOCK_SIZE;
    if (byteOffset < 0) {
      ASSERT (request->frameNumber > 0);
      ASSERT (BLOCK_SIZE == 512);
      //      ASSERT (blocksPerFrame == 128);
      ASSERT (byteOffset >= 0);
    }
    totalSeekDistance += abs(request->frameNumber - head);  

    pthread_mutex_lock(&requestQueue->mutex);
    head = request->frameNumber + 1;
    pthread_mutex_unlock(&requestQueue->mutex);
#ifdef DEBUG_DISK
    cout << byteOffset << " byteOffset " << endl;
#endif

    // just leave the first block empty
    if (lseek(fd, byteOffset + BLOCK_SIZE, SEEK_SET) != byteOffset + BLOCK_SIZE) {
      perror ("Disk::doWrite (lseek):");
      return DISK_WRITE_ERROR;
    }
  } else {
    pthread_mutex_lock(&requestQueue->mutex);
    head = request->frameNumber + 1;
    pthread_mutex_unlock(&requestQueue->mutex);
  }


  if (write(fd, request->data, frameSize) < 0) {
    perror ("Disk::doWrite (write):");    
    return DISK_WRITE_ERROR;
  }

  return OK;
}

//=============================================================================
Status Disk::doRead (Request *request) {
#ifdef DEBUG_DISK
  cout << "Starting read request " << request << endl;
#endif

  // set the disk head to the right location on disk

  if (head != request->frameNumber)  {
    int byteOffset = request->frameNumber * CFS::blocksPerFrame * BLOCK_SIZE;
    if (byteOffset < 0) {
      ASSERT (request->frameNumber > 0);
      ASSERT (BLOCK_SIZE == 512);
      //      ASSERT (blocksPerFrame == 128);
      ASSERT (byteOffset >= 0);
    }
    totalSeekDistance += abs(request->frameNumber - head);

    // move the head first so that reads which come in during the disk access
    // will be positioned correctly.
    pthread_mutex_lock(&requestQueue->mutex);
    head = request->frameNumber + 1;
    pthread_mutex_unlock(&requestQueue->mutex);

    if (lseek(fd, byteOffset + BLOCK_SIZE, SEEK_SET) != byteOffset + BLOCK_SIZE) {
      perror ("Disk::doRead:");
      return DISK_READ_ERROR;
    }
  } else {
    pthread_mutex_lock(&requestQueue->mutex);
    head = request->frameNumber + 1;
    pthread_mutex_unlock(&requestQueue->mutex);
  }

  // read the data off of the disk
  if(read(fd, request->data, frameSize) != frameSize) {
    perror ("Disk::doRead:");    
    return DISK_READ_ERROR;
  }

  return OK;
}

//=============================================================================
ostream& operator << (ostream& os, Disk& disk) {
  os << "Total seek distance=" << disk.totalSeekDistance << endl;
  os << "Max request queue=" << disk.maxRequestQueue << endl;
  return os;
}
