/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ffs.h"

FFS::FFS  (const char *const base, Status& returnStatus) : baseDirectory (base) {
  //  this->baseDirectory = baseDirectory;
  filesRead = 0;
  filesWritten = 0;
  cout << "ffs ctor" << endl;
  returnStatus = OK;
}

FFS::~FFS  () {
  cout << "ffs destructor" << endl;
}

Status FFS::lookup (const char *const url, int & groupNumber) {
  FileDesc desc;
  genDesc (url, desc);

  groupNumber = 0;

#ifdef DEBUG_LOOKUP
  cout << "Lookup: " << desc.fileName << endl;
#endif

  struct stat buf;

  if ( (::stat (desc.fileName, &buf)) == 0)
    return OK;
  return FILE_NOT_FOUND;
}

Status FFS::create (const char *const url, const int group, const char *const data, const int size) {
  FileDesc desc;
  genDesc (url, desc);

#ifdef DEBUG_CREATE
  cout << "Create: " << desc.fileName << endl;
#endif

  int fd, error;

  // create the file
  if ((fd = ::open (desc.fileName, O_WRONLY | O_CREAT | O_TRUNC, S_IWUSR | S_IRUSR )) < 0) {
    cerr << "Error creating file: " << desc.fileName << endl;
    return IO_ERROR;
  }

  // write the data to the file
  if((error = ::write(fd, data, size)) < 0){
    cerr << "Error writing to file, fd="<< fd << " name=" << desc.fileName << endl;
    return IO_ERROR;
  }

  // close the file
  if((error = ::close(fd)) < 0){
    cerr << "Error closing file.\n";
    return IO_ERROR;
  }

  filesWritten++;
  return OK;
}

Status FFS::read (const char *const url, const int groupNumber, char *&data, int &length) {
  int fd, error;
  FileDesc desc;
  genDesc (url, desc);

#ifdef DEBUG_READ
  cout << "Read: " << desc.fileName << endl;
#endif

  if((fd = ::open(desc.fileName, O_RDONLY)) < 0){
    return IO_ERROR;
  }

  struct stat buf;

  if ( (::fstat (fd, &buf)) < 0) {
    cerr << "Error fstat-ing file.\n";
    return IO_ERROR;
  }

  data = new char [buf.st_size];
  if (data == NULL) {
    return OUT_OF_MEMORY;
  }
  length = buf.st_size;

  // read the files data into memory
  if((error = ::read(fd, data, buf.st_size)) < 0){
    cerr << "Error reading file.\n";
    return IO_ERROR;
  }

  // close the file
  if((error = ::close(fd)) < 0){
    cerr << "Error reading file.\n";
    return IO_ERROR;
  }

  filesRead++;
  return OK;
}

Status FFS::close (const char *const url) {

  return OK;
}

Status FFS::remove (const char *const url, int groupNumber) {
  FileDesc desc;
  genDesc (url, desc);
  if (unlink(desc.fileName) < 0){
    cerr << "Error unlinking the file.\n";
    return IO_ERROR;
  }
  return OK;
}

void FFS::genDesc (const char *url, FileDesc &desc) { 
  //  md5_byte_t *digest = new md5_byte_t[16];
  genMD5 (url, desc.digest);
  desc.dirA = (char)(97 + (desc.digest[0] % 26));
  desc.dirB = (char)(97 + (desc.digest[1] % 26));
  //  cout << "(A, B) = (" << desc.dirA << "," << desc.dirB << ")" << endl;
  //  printMD5 (desc.digest);

  // char *fileName = new char (strlen (baseDirectory) + sizeof (FileDesc) + 2);

  memset (desc.fileName, 0, sizeof (desc.fileName));
  strcpy (desc.fileName, baseDirectory);
  int fileNameIndex = strlen (baseDirectory);
  desc.fileName [fileNameIndex++] = '/';
  desc.fileName [fileNameIndex++] = desc.dirA;
  desc.fileName [fileNameIndex++] = '/';
  desc.fileName [fileNameIndex++] = desc.dirB;
  desc.fileName [fileNameIndex++] = '/';
  for (int di = 0; di < 16; ++di) {
     sprintf (&desc.fileName[fileNameIndex], "%02x", desc.digest[di]);
     fileNameIndex += 2;
  }
   //   printf("\n");
   //  strcpy (fileName + strlen (baseDirectory) + 4, (char *) desc.digest);

   return;
}

void FFS::genMD5 (const char *url, md5_byte_t *digest) {
  md5_state_t state;
  //        md5_byte_t digest[16];

  md5_init(&state);
  md5_append(&state, (const md5_byte_t *)url, strlen(url));
  md5_finish(&state, digest);

//   printf("MD5 (\"%s\") =\n", url);
//   for (int di = 0; di < 16; ++di)
//     printf("%02x", digest[di]);
//   printf("\n");
  return;
}

void FFS::printMD5 (md5_byte_t *digest) {
  int di;
  cout << "size = " << sizeof (md5_byte_t *) << endl;

  for (di = 0; di < 16; ++di)
    printf("%02x", digest[di]);
  printf("\n");
}

ostream& operator << (ostream& os, FFS& ffs) 
{
  os << "File system:";
  os << "\n\treads: " << ffs.filesRead;
  os << "\n\twrites: " << ffs.filesWritten;
  return os;
}
      
