/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "bufferManager.h"
#include "file.h"

File::File (const int myID, RequestQueue *theQueue, const int off, const int fcount) : id (myID), queue (theQueue), offset (off), maxFrames (fcount) {
  numFrames = 0;

  deleteQueue = new DeleteQueue (maxFrames);

  // no page number 0
  //  pagesInMemory.insert (pagesInMemory.begin(), 0);

#ifdef DEBUG_FILE
  cout << "File begins at offset " << offset << " and has " << maxFrames << " frames." << endl;
#endif 
  

}

//=======================================================================
File::~File () {
  //  cout << "A deleting file" << endl;
  delete deleteQueue;
  //  cout << "B deleting file" << endl;
}

//=======================================================================
const Status File::writePage (const int pageNo, Page* pagePtr) {
 if (pageNo < 0 || pageNo > numFrames) {
    return BADPAGENO;
  }

 Status status;

 // pthread_mutex_lock (&mutex); 

 // int blockOffset = offset + (pageNo * (frameSize / BLOCK_SIZE));

 int frameNumber = offset + pageNo;

 ASSERT (pagePtr >=  BufMgr::lowFrameAddress && pagePtr <=  BufMgr::highFrameAddress);
 Request *request = new Request (WRITE, pagePtr, frameNumber);
// #ifdef DEBUG_FILE
//  cout << request << endl;
// #endif

 pthread_mutex_lock (&request->mutex);

 pthread_mutex_lock (&queue->mutex);

#ifdef DEBUG_QUEUE
 ASSERT ( (status = queue->verify ()) == OK);
#endif

 if ( (status = queue->enqueue (request)) != OK) {
   pthread_mutex_unlock (&request->mutex);
   return status;
 }

#ifdef DEBUG_QUEUE
 ASSERT ( (status = queue->verify ()) == OK);
#endif

 pthread_mutex_unlock (&queue->mutex);

 while (!request->complete) {
   pthread_cond_wait (&request->cv, &request->mutex);
 }

 pthread_mutex_unlock (&request->mutex);

#ifdef USE_DELETE_QUEUE
 // remove any pending deletes to this page before passing it up to the buffer manager
 pthread_mutex_lock (&deleteQueue->mutex);
 RID rid;
 // cout << "id" << id << endl;
 while (deleteQueue->dequeue (pageNo, rid) == OK && rid.slotNo >= 0) {
   //   cout << "deleting " << rid << endl;
   status = pagePtr->deleteRecord (rid);
   if (status != OK && status != NORECORDS) {
     pthread_mutex_unlock (&deleteQueue->mutex);
     return status;
   }
 }
 pthread_mutex_unlock (&deleteQueue->mutex);
#endif


  bool found = false;
  for (unsigned int i = 0; i < pagesInMemory.size() && !found; i++) {
    // cout << "pagesInMemory[" << i << "] = " << pagesInMemory[i] << endl;
    if (pagesInMemory[i] == pageNo) {
      pagesInMemory.erase (pagesInMemory.begin() + i);
      found = true;
    }
  }
  if (found == false) {
    return PAGES_IN_MEMORY_VECTOR_ERROR;
  }

  //  pthread_mutex_unlock (&mutex); 
  return OK;
}

//=======================================================================
const Status File::readPage (const int pageNo, Page* pagePtr) {
  if (pageNo < 0 || pageNo > numFrames) {
    return BADPAGENO;
  }

  Status status;

  // pthread_mutex_lock (&mutex); 

  //  int frameOffset = offset + (pageNo * (frameSize / BLOCK_SIZE));

 int frameNumber = offset + pageNo;
 ASSERT (pagePtr >=  BufMgr::lowFrameAddress && pagePtr <= BufMgr::highFrameAddress);
 Request *request = new Request (READ, pagePtr, frameNumber);
// #ifdef DEBUG_FILE
//  cout << request << endl;
// #endif

 pthread_mutex_lock (&request->mutex);

 pthread_mutex_lock (&queue->mutex);

#ifdef DEBUG_QUEUE
 ASSERT ( (status = queue->verify ()) == OK);
#endif


 if ( (status = queue->enqueue (request)) != OK) {
   pthread_mutex_unlock (&request->mutex);
   return status;
 }

#ifdef DEBUG_QUEUE
 ASSERT ( (status = queue->verify ()) == OK);
#endif

 pthread_mutex_unlock (&queue->mutex);

 while (!request->complete) {
   pthread_cond_wait (&request->cv, &request->mutex);
 }

 pthread_mutex_unlock (&request->mutex);

#ifdef USE_DELETE_QUEUE
 // remove any pending deletes to this page before passing it up to the buffer manager
 pthread_mutex_lock (&deleteQueue->mutex);
 RID rid;
 while (deleteQueue->dequeue (pageNo, rid) == OK && rid.slotNo >= 0) {
   status = pagePtr->deleteRecord (rid);
   if (status != OK && status != NORECORDS) {
     pthread_mutex_unlock (&deleteQueue->mutex);
     return status;
   }
 }
 pthread_mutex_unlock (&deleteQueue->mutex);
#endif

 pagesInMemory.insert (pagesInMemory.end(), pageNo);

  //  pthread_mutex_unlock (&mutex); 
  return OK;
}

//=======================================================================
const Status File::allocatePage (int& pageNo) {
  //  pthread_mutex_lock (&mutex); 
  numFrames++;
  if (numFrames > maxFrames) {
    return DISK_FULL;
  }

#ifdef DEBUG_FILE
  //  if (numFrames % 10 == 0) 
    cout << pthread_self() << " allocated " << numFrames << " frames in file " << id << endl;
#endif

  pageNo = numFrames - 1;
  pagesInMemory.insert (pagesInMemory.end(), pageNo);
  // pthread_mutex_unlock (&mutex); 
  return OK;
}

//=======================================================================
ostream& operator << (ostream& os, File &file) {
  os << file.id;
  return os;
}

//=======================================================================
DeleteQueue::DeleteQueue (const int fCount) : frameCount (fCount) {
  pthread_mutex_init (&mutex, NULL);

  queue = new RIDnode* [frameCount];
  for (int i = 0; i < frameCount; i++) {
    queue[i] = NULL;
  }

}

//=======================================================================
DeleteQueue::~DeleteQueue () {

}

//=======================================================================
Status DeleteQueue::enqueue (RID rid) {
  pthread_mutex_lock (&mutex);
  RIDnode *newNode = new RIDnode (rid);

  // queue for this page was empty
  if (queue [rid.pageNo] == NULL) {
    queue [rid.pageNo] = newNode;
    pthread_mutex_unlock (&mutex);
    return OK;
  }

  RIDnode *current = queue [rid.pageNo];

  if (rid.slotNo > current->slotNo) {
    // put it at the front
    newNode->next = current;
    queue[rid.pageNo] = newNode;
    pthread_mutex_unlock (&mutex);
    return OK;
  }

  while (current->next != NULL) {
    if (current->slotNo  > rid.slotNo && current->next->slotNo < rid.slotNo) {
      newNode->next = current->next;
      current->next = newNode;
      pthread_mutex_unlock (&mutex);
      return OK;
    }
    current = current->next;
  }

  // put it at the end
  current->next = newNode;
  pthread_mutex_unlock (&mutex);
  return OK;
}

//=======================================================================
// this lock surrounds the dequeue call
// this is because we anticipate making repeated dequeue calls,
// not so with enqueue
Status DeleteQueue::dequeue (int pageNo, RID &rid) {
  //  RID rid;
  rid.pageNo = pageNo;

  if (queue[pageNo] == NULL) {
    rid.slotNo = -1;
    return OK;
  }
  rid.slotNo = queue[pageNo]->slotNo;
  RIDnode *front = queue[pageNo];
  queue[pageNo] = front->next;
  delete front;
  return OK;

}

//=============================================================================
ostream& operator << (ostream& os, DeleteQueue& queue) {
  for (int i = 0; i < queue.frameCount; i++) {
    if (queue.queue[i] != NULL) {
      os << i << " ";
      RIDnode *current = queue.queue [i];
      while (current->next != NULL) {
	os << "(" << current->pageNo << "," << current->slotNo << ") ";
	current = current->next;
      }
      os << "(" << current->pageNo << "," << current->slotNo << ") ";
      os << endl;
    }
  }
  return os;
}


//=======================================================================
RIDnode::RIDnode (RID rid) {
  this->pageNo = rid.pageNo;
  this->slotNo = rid.slotNo;
  this->next = NULL;
}

RIDnode::~RIDnode () {
}
