/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include "group.h"
#include "cfs.h"

Group::Group (int myGroupNumber, BufMgr* myBufMgr, RequestQueue *queue, int offset, int diskFramesAvailable, Status& returnStatus) : bufMgr (myBufMgr), groupNumber (myGroupNumber)  {

  pthread_mutex_init (&allocMutex, NULL);
 // no page number 0
  // freeSpace.insert (freeSpace.begin(), 0);
  loveIt = false;
  pageCount = 0;
  recordCount = 0;

  lastPageNo = 0;

  fileTable = new FileTable ();

  //  char baseFileName[100];
  //  sprintf (baseFileName, "/home/ledlie/proxy/cfs_cache/group.%d", groupNumber);
  //  sprintf (baseFileName, "/u/l/e/ledlie/private/thesis/cfs_cache/group.%d", groupNumber);

  //  fileName = baseFileName;

  cout << "Group\t" << groupNumber << " offset\t" << offset << "\tdisk frames:\t" << diskFramesAvailable << endl; 

  file = new File (groupNumber, queue, offset, diskFramesAvailable);

  pagesInMemory = &(file->pagesInMemory);

  returnStatus = OK;

#ifdef DEBUG_GROUP
  cout << "Created group " <<groupNumber << endl;
#endif

}

//=============================================================================
Group::~Group () {

  //  still need to write fileTable out to disk
  //  cout << *fileTable << endl;
#ifdef DEBUG_GROUP
  cout << *this << endl;
#endif

  //  cout << "deleting group" << endl;

  delete fileTable;
  delete file;  

}

//=============================================================================
// insert record into group
const Status Group::insertRecord(const  md5_byte_t * const digest, const Record rec) {

#ifdef DEBUG_CREATE
  static int timesThrough;



  cout << "top of Group(" << groupNumber << ")::insertRecord (x" << timesThrough++ << ")" << endl;
#endif

  Status status;
  RID rid;
  int looping = 0;

  // make sure file doesn't exist already
  status = fileTable->lookup (digest, rid);
  if (status != FILE_NOT_FOUND) {
#ifdef DEBUG_CREATE
    cout << "Lookup failure in Group::create(): " << status << endl;
#endif
    return status;
  }

#ifdef DEBUG_CREATE
    cout << "Returned from lookup in Group::create(): " << status << endl;
#endif

  status = NOSPACE;
  int pageNo = -1;
  Page *page;
  vector<int> pagesAttempted;
  bool forcePageAlloc = false;

  while (status == NOSPACE) {
    looping++;
    
    if (looping > 4 ) {
      //      for (unsigned int i = 0; i < pagesAttempted.size(); i++) {
      //	cout << pagesAttempted[i] << "(fs: " << freeSpace[pagesAttempted[i]] << ")\t";
	//      }
      //      cout << "\n" << pthread_self() << "\nlooped too many times in Group::create(), length:" << rec.length << endl;
      //      exit (0);
      forcePageAlloc = true;
    }

    // figure out which pageNo has enough space for this record
    //    int oldPageNo = pageNo;
    status = findFreeBuffer (rec.length, pageNo, page, pagesAttempted, forcePageAlloc);

    // if (status == NEW_BUFFER) {
//       cout << "created a new buffer" << endl;
//     }

    pagesAttempted.insert (pagesAttempted.end(), pageNo);

    // status can be OK or NEW_BUFFER (or bad)
    if (status != NEW_BUFFER) {
      //  cout << "did not create a new buffer" << endl;
      status = bufMgr->readPage (file, pageNo, page, EXCLUSIVE);
      if (status != OK) {
	return status;
      }
    }
#ifdef DEBUG_CREATE
    cout << "trying to insert record of size " << rec.length << " into page number " << pageNo << " which has " << page->getFreeSpace() << " bytes free" << endl;
#endif

    status = page->insertRecord (rec, rid);

    // cout << "inside create\n" << *bufMgr << endl;

    //   cout << "about to unpin page, total now: " << bufMgr->getPinnedPageCount() << endl;
    // ASSERT (bufMgr->getPinnedPageCount() == 1);

    freeSpace[pageNo] = page->getFreeSpace ();
    Status unPinStatus =  bufMgr->unPinPage (file, pageNo, DIRTY, loveIt, EXCLUSIVE);
    if (unPinStatus != OK) return unPinStatus;

    //  cout << "unpinned page, total now: " << bufMgr->getPinnedPageCount() << endl;

  }
  if (status != OK) { 
#ifdef DEBUG_CREATE
    cout << status << endl;
#endif
    return status;
  }

#ifdef DEBUG_CREATE
  cout << "inserted rid: " << rid << endl;
#endif

  recordCount++;
  fileTable->insert (digest, rid);

#ifdef DEBUG_CREATE
  cout << "bottom create: " << rid << endl;
#endif
  

  return OK;
}

#ifdef USE_DELETE_QUEUE
const Status Group::removeRecord(const  md5_byte_t * const digest) {
  Status status;
  RID rid;
  status = fileTable->lookup (digest, rid);
  if (status != OK) return status;

  status = fileTable->remove (digest);
  if (status != OK) return status;

  // if it's in memory, mark it dirty
  // no locking necessary.
  // if we accidentally mark a clean page dirty (which there is a very small change of happening,
  // no real loss, except for one i/o
  bool found = false;
  for (unsigned int i = 0; !found && i < pagesInMemory->size(); i++) {
    int tmpPageNo = (*pagesInMemory)[i];
    if (tmpPageNo == rid.pageNo) {
      found = true;
      status = bufMgr->markDirty (file, rid.pageNo);
      if (status != OK) return status;
    }
  }
  status = file->deleteQueue->enqueue (rid);
  return status;
}

#else

//=============================================================================
// delete record from group
const Status Group::removeRecord(const  md5_byte_t * const digest) {

#ifdef DEBUG_DELETE
  cout << "top of Group::removeRecord" << endl;
#endif

  Status status;
  RID rid;
  status = fileTable->lookup (digest, rid);
  if (status != OK) return status;

  Page *page;
  status = bufMgr->readPage (file, rid.pageNo, page, EXCLUSIVE);
  if (status != OK) return status;

  //  cout << "got here 1" << endl;

  status = page->deleteRecord (rid);
  if (status != OK && status != NORECORDS) {
    cout << status << endl;
    return status;
  }

  // set free space
  freeSpace[rid.pageNo] = page->getFreeSpace ();
  recordCount--;

  status = bufMgr->unPinPage (file, rid.pageNo, DIRTY, loveIt, EXCLUSIVE);
  if (status != OK) return status;

  //  cout << "got here 3" << endl;

  status = fileTable->remove (digest);
  if (status != OK) return status;

  return OK;
}

#endif

//=============================================================================
const Status Group::lookupRecord(const  md5_byte_t * const digest) {
#ifdef DEBUG_GROUP
  //  cout << "top of Group::lookupRecord" << endl;
#endif

  RID rid;
  Status status;
  status = fileTable->lookup (digest, rid);
  //  cout << status << endl;
  return status;
}


//=============================================================================
// returns a pointer by reference to a record in a buffer
// the page where this record is is pinned read-only
const Status Group::readRecord(const  md5_byte_t * const digest, Record & rec) {

#ifdef DEBUG_READ
  cout << "top of Group::readRecord" << endl;
#endif

  Status status;
  RID rid;
  status = fileTable->lookup (digest, rid);
  if (status != OK) return status;

  Page *page;
  status = bufMgr->readPage (file, rid.pageNo, page, SHARED);
  if (status != OK) return status;

  status = page->getRecord (rid, rec);
  if (status != OK) return status;

  return OK;
}

//=============================================================================
// unpins the page which is holding this record
const Status Group::releaseRecord(const  md5_byte_t * const digest) {

  Status status;
  RID rid;
  status = fileTable->lookup (digest, rid);
  if (status != OK) return status;

  status = bufMgr->unPinPage (file, rid.pageNo, CLEAN, loveIt, SHARED);
  if (status != OK) return status;

  return OK;

}

//=============================================================================
const int Group::getRecordCount() const {
  return recordCount;
}

//=============================================================================
const Status Group::flush () {
  // have disk locations as vars?
  return OK;

}

//=============================================================================
const Status Group::setLoveIt (bool myLoveIt) {
  loveIt = myLoveIt;
  return OK;
}

//=============================================================================
// if pageNo is passed in, then we will return a different pageNo.
// this happens when insertRecord returns NOSPACE
const Status Group::findFreeBuffer (int length, int &pageNo, Page * &newPage, vector<int> &pagesAttempted, bool forcePageAlloc) {

  bool beenHere = false;

#ifdef FIND_FREE_BUFFER
  do {

    // put in a pause here if we've gone through before and some other thread is presently allocating
    // a buffer for this group
    if (beenHere) {
      pthread_mutex_lock (&allocMutex);
      pthread_mutex_unlock (&allocMutex);
    }

#else
  pthread_mutex_lock (&allocMutex);  
#endif

  ASSERT ((unsigned)pageCount == freeSpace.size());

  if (!forcePageAlloc) {

  // is there an available page in the buffer pool already?
  for (unsigned int i = 0; i < pagesInMemory->size(); i++) {

    int tmpPageNo = (*pagesInMemory)[i];

    if (freeSpace [tmpPageNo] > length) {
      bool found = false;
      for (unsigned int j = 0; j < pagesAttempted.size() && !found; j++) {
	if (pagesAttempted[j] == tmpPageNo) {
	  found = true;
	}
      }
      if (!found) {
	pageNo = tmpPageNo;
#ifndef FIND_FREE_BUFFER
	pthread_mutex_unlock (&allocMutex);  
#endif
	return OK;
      }
    }
  }

  
  // is there a page on disk with enough space?
  for (unsigned int i = 0; i < freeSpace.size(); i++) {
    if (freeSpace[i] > length) {
#ifdef DEBUG_CREATE
      cout << "pageNo: " << i << " has " << freeSpace[i] << " bytes free.  Enough for " << length << endl;
#endif
      bool found = false;
      for (unsigned int j = 0; j < pagesAttempted.size() && !found; j++) {
	if ((unsigned)pagesAttempted[j] == i) {
	  found = true;
	}
      }
      if (!found) {
	pageNo = i;
#ifndef FIND_FREE_BUFFER
	pthread_mutex_unlock (&allocMutex);  
#endif
	return OK;
      }
    }
  }

  } // end of force alloc

#ifdef FIND_FREE_BUFFER
  beenHere = true;
  } while (pthread_mutex_trylock (&allocMutex) != 0);
    // we got the lock, meaning that no one else is allocating a buffer right now
#endif


  // we've got to make a new page
  //  Page *newPage;
  Status status;
  status = bufMgr->allocPage (file, pageNo, newPage);
  if (status != OK) {
#ifdef DEBUG_CREATE
    cout << "Could not allocate a page in Group::findFreeBuffer" << endl;
#endif
    long int totalFreeSpace = 0;
    long int totalSpace = 0;
    for (unsigned int i = 0; i < freeSpace.size(); i++) {
      totalFreeSpace += freeSpace[i];
      totalSpace += frameSize;
      cout << i << "\t" << freeSpace[i] << endl;
    }
    cout << "Free:\t" << totalFreeSpace << endl;
    cout << "Total:\t" << totalSpace << endl;

    ASSERTSTATUS (false);
    pthread_mutex_unlock (&allocMutex);  
    return status;
  }
  newPage->init (pageNo);

  int amt = newPage->getFreeSpace();

  // assuming that pageNo's are allocated in order
  freeSpace.insert (freeSpace.end(), amt);

  lastPageNo++;
  pageCount++;

  pthread_mutex_unlock (&allocMutex);  
  return NEW_BUFFER;

}

//=============================================================================
ostream& operator << (ostream& os, Group& group) {
  os << "Group " << group.groupNumber << "\n\t";
  if (group.bufMgr == NULL) {
    os << "Null buffer manager\n\t";
  }
  // free space array
  //os << "Free space: TBA\n\t";

  // which pages are in memory
  //  os << "Pages in memory: TBA\n\t";

  if (group.loveIt) {
    os << "\n\tLoved.\n";
  } else {
    os << "\n\tHated.\n";
  }
  os << "\tPage count: " << group.pageCount << "\n\tRecord count: " << group.recordCount << "\n";

  cout << *group.fileTable << endl;
  return os;
}
