/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GROUP_H
#define GROUP_H

#include <utility>
#include <vector>
#include <hash_map>
#include "fileTable.h"
#include "md5.h"
#include "requestQueue.h"
#include "page.h"
#include "status.h"
#include "globals.h"
#include "bufferManager.h"
#include "file.h"

class Group {
  friend ostream& operator << (ostream& os, Group& group);
 public:
  Group (int groupNumber, BufMgr* bufMgr, RequestQueue *queue, int offset, int diskFramesAvailable, Status& returnStatus);
  ~Group ();

  // insert record into group
  const Status insertRecord(const  md5_byte_t * const digest, const Record rec); 

  // delete record from group
  const Status removeRecord(const  md5_byte_t * const digest);

  //  const Status lookupRecord(const char *digest);
  const Status lookupRecord(const  md5_byte_t * const digest);

  // returns a pointer by reference to a record in a buffer
  // the page where this record is is pinned read-only
  const Status readRecord(const  md5_byte_t * const digest, Record & rec);

  // unpins the page which is holding this record
  const Status releaseRecord(const  md5_byte_t * const digest);

  const int getRecordCount() const;
  const Status flush (); // have locations as vars?

  const Status setLoveIt (bool loveIt);

  // public for debug
  // every Group has one instance of a FileTable
  FileTable *fileTable;

#ifdef QUICK_CFS_LOOKUP
  FileTable* getFileTable () { return fileTable; }
#endif

 private:

  // for first development stage, every group has pointer to File object (which is a real file)
  File *file;

  const string baseDirectory;

  // just two pointers to keep things private
  BufMgr* bufMgr;

  const int groupNumber;

  string fileName;

  // mutex to prevent a page from allocating more than one page at a time
  pthread_mutex_t allocMutex;

  // vector where the keys are page numbers and the values are the amount of free space on that page
  vector<int> freeSpace;
  // array of the page numbers which are currently in memory 
  // (actually part of file object because it know when things go to disk)
  vector<int> *pagesInMemory;
  const Status findFreeBuffer (int length, int &pageNo, Page * &page, vector<int> &pagesAttempted, bool forcePageAlloc);

  // true if unpinned pages from this group go on the rear of the replacement queue
  bool loveIt;

  // number of pages in this group
  int pageCount;

  int lastPageNo;

  // number of records
  int recordCount;

};

#endif
