/*
 * Copyright (C) 2001  Jonathan T. Ledlie
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "bufferManager.h"
#include "cfs.h"
#include "requestQueue.h"

#ifdef DEBUG_QUEUE
#define DEBUG_QUEUE_OUT(c) cout << c << endl;
#define CHECKFRAME(c)     { if (c->frameNumber == request->frameNumber) { \
                            cout << "In queue: " << c << "\nNew request: " << request << endl; \
                            int offset = c->frameNumber % CFS::framesPerGroup; \
	                    int groupNo = c->frameNumber / CFS::framesPerGroup; \
	                    cout << "(" << groupNo << "," << offset << ") "; \
                            BufMgr::printReadList (); \
                            cout << *this << endl; } \
                            ASSERT (c->frameNumber != request->frameNumber); \
                          }
#else
#define DEBUG_QUEUE_OUT(c)
#define CHECKFRAME(c)
#endif


RequestQueue::RequestQueue (int *diskHead) : head (diskHead) {
  pthread_mutex_init (&mutex, NULL);
  pthread_cond_init (&cv, NULL);
  first = NULL;

#ifdef FIFO
  last = NULL;
  cout << "Request Queue using FIFO" << endl;
#else
  least = NULL;
  cout << "Request Queue using One Way Elevator" << endl;
#endif
  length = 0;

}

RequestQueue::~RequestQueue () {
}

#ifdef FIFO

//=============================================================================
// FIFO version of enqueue
// enter and exit here with requestQueue's lock
Status RequestQueue::enqueue (Request *request) {

  Status status;
  static int duplicates;
  static int timesThrough;

  DEBUG_QUEUE_OUT ("Queuing " << request);

  // queue empty when request arrives
  if (first == NULL){
    first = request;
    last = request;
    length++;

    // wake up the disk
    pthread_cond_signal (&cv);
    return OK;
  }

  timesThrough++;

  bool duplicate = false;

  // check for duplicate (happens rarely)
  // should only be write request in queue and
  // current is read
  Request *current = first;
  while (current->next != NULL) {
    if (current->frameNumber == request->frameNumber) {
      DEBUG_QUEUE_OUT ("current: " << current->frameNumber << " request: " << request->frameNumber);
      ASSERT (current->action == WRITE && request->action == READ);
      //   pause ();
      duplicates++;
      DEBUG_QUEUE_OUT ("*************Duplicates " << duplicates << " out of " << timesThrough << " (when queue is not empty)");
      request->complete = true;
      memcpy ((char*)request->data, (char*)current->data, frameSize);
      return OK;
    }
    duplicate = true;
    current = current->next;
  }
    if (current->frameNumber == request->frameNumber) {
      DEBUG_QUEUE_OUT ("current: " << current->frameNumber << " request: " << request->frameNumber);
      ASSERT (current->action == WRITE && request->action == READ);
      //    pause ();
      duplicates++;
      DEBUG_QUEUE_OUT ("*************Duplicates " << duplicates << " out of " << timesThrough << " (when queue is not empty)");
      duplicate = true;
      request->complete = true;
      memcpy ((char*)request->data, (char*)current->data, frameSize);
      return OK;
    }
    

  // no duplicate, do regular insertion at end

  // queue is not empty, put it at the end
  current = first;
  while (current->next != NULL) {
    current = current->next;
  }
  current->next = request;
  request->previous = current;
  last = request;
  length++;

  if (request->next != NULL) {
    return REQUEST_QUEUE_ERROR;
  }

  return OK;
}

//=============================================================================
// enter and exit here with lock on requestQueue
Status RequestQueue::dequeue (Request* &firstRequest) {
  firstRequest = first;

  // only one entry in the queue
  if(first->next == NULL){
    first = NULL;
    last = NULL;
  }
  else
    first = first->next;
  length--;

  // if the queue is not empty, the first item better not be null, and visa versa
  ASSERT ( (length > 0 && first != NULL) || (length == 0 && first == NULL));

  DEBUG_QUEUE_OUT ("Dequeuing " << firstRequest);

  return OK;
}

ostream& operator << (ostream& os, RequestQueue& queue) {
  Request *current = queue.first;
  while (current != NULL) {
    os << current->frameNumber << " ";
    current = current->next;
  }
  os << endl;
  return os;
}

#else

//=============================================================================
// Simple Elevator version of enqueue
// enter and exit here with requestQueue's lock

Status RequestQueue::enqueue (Request *request) {

  // Status status;
  static int duplicates;
  static int timesThrough;

  //  DEBUG_QUEUE_OUT ("Length = " << length << ", Head = " << *head << " request->frameNum = " << request->frameNumber);

  // queue empty when request arrives
  if (first == NULL){
    first = request;
    least = request;
    request->next = request;
    request->previous = request;

    length++;

    // wake up the disk
    pthread_cond_signal (&cv);
    return OK;
  }

  timesThrough++;

  // check for duplicate (happens rarely)
  // should only be write request in queue and
  // current is read

  // doh!  it can happen for two write request
  // in this case, we dequeue the first (outdated) write and
  // insert the new one in its place
  // we then wake up the first's requester

  Request *duplicateRequest = least;

  int loop = 0;

  while (loop < length) {

    loop++;

    if (loop++ > 500) {
      cout << "Deadlock!" << endl;
      pause ();
    }

    if (duplicateRequest->frameNumber == request->frameNumber) {
      DEBUG_QUEUE_OUT ("duplicateRequest: " << duplicateRequest->frameNumber << " request: " << request->frameNumber);




      if (duplicateRequest->action == WRITE && request->action == READ) {
	request->complete = true;
	memcpy ((char*)request->data, (char*)duplicateRequest->data, frameSize);

      } else if (duplicateRequest->action == WRITE && request->action == WRITE) {

	DEBUG_QUEUE_OUT ("********************************* Two writes");

	// remove old (duplicate) request and put new one in it's place
	// we do this (instead of just marking the new one as done)
	// in order to keep the locks on the requests straight

	request->next = duplicateRequest->next;
	request->previous = duplicateRequest->previous;

	request->next->previous = request;
	request->previous->next = request;
	
	duplicateRequest->complete = true;

	if (pthread_mutex_trylock(&duplicateRequest->mutex) == 0) {
	  pthread_cond_signal (&duplicateRequest->cv);
	  pthread_mutex_unlock(&duplicateRequest->mutex);
	}

      } else {

	// cannot have two readers -- this should be prevented with readList in bufferManager
#ifdef DEBUG_QUEUE	
	cout << "on-list action: ";
	if (duplicateRequest->action == WRITE) cout << "w"; else cout << "r";
	cout <<  ", new request action: ";
	if (request->action == WRITE) cout << "w"; else cout << "r";
	cout << endl;
#endif
	ASSERT (false);
      }

      duplicates++;
      DEBUG_QUEUE_OUT ("**Duplicates " << duplicates << " out of " << timesThrough << " (when queue is not empty)");
      return OK;
    }
    //duplicate = true;
    duplicateRequest = duplicateRequest->next;
  }




  // one member of queue
  if (least->previous == least) {
    // cout << length << endl;
    ASSERT (length == 1);
    CHECKFRAME (least);

    request->next = least;
    request->previous = least;
    least->previous = request;
    least->next = request;
    if (request->frameNumber < least->frameNumber) {
      least = request;
    }

  } else {
    // two or more requests already on queue
    ASSERT (length > 1);

    // see if we have a new lowest frame number
    if (request->frameNumber < least->frameNumber) {
      CHECKFRAME (least);
      request->next = least;
      request->previous = least->previous;
      least->previous->next = request;
      least->previous = request;
      least = request;
    } else {
      // the request is greater than the least 
      // and may be greater than the largest current list member

      Request *current = least;
      bool found = false;
      while (!found && current->next != least) {
	CHECKFRAME (current);
	// error condition

// 	if (current->frameNumber == request->frameNumber) {


// 	  ASSERT (false);
// 	}

	if (current->frameNumber < request->frameNumber && current->next->frameNumber > request->frameNumber) {
	  found = true;
	} else {
	  current = current->next;
	}
      }
      CHECKFRAME (current);
      //      ASSERT (current->frameNumber != request->frameNumber);

      request->next = current->next;
      request->previous = current;
      current->next->previous = request;
      current->next = request;
     
    }
  }

  length++;
  // now that we've inserted this request into the queue
  // reposition the first pointer

  first = NULL;
  if (*head > least->previous->frameNumber || *head <= least->frameNumber) {
    first = least;
    //    cout << "setting first to " << least->frameNumber << endl;
  } else {
    Request *current = least;
    bool found = false;
    while (!found && current->next != least) {
      if (current->frameNumber <= *head && current->next->frameNumber > *head) {
	found = true;
      } else {
	current = current->next;
      }
    }
    first = current;
    //    cout << "setting first to " << current->frameNumber << endl;

  }


  ASSERT (first != NULL);


  return OK;
}

//=============================================================================
// enter and exit here with lock on requestQueue
Status RequestQueue::dequeue (Request* &firstRequest) {
  firstRequest = first;

  if (first == NULL)
    return REQUEST_QUEUE_ERROR;

  // only one entry in the queue
  if(first->next == first){
    first = NULL;
    least = NULL;
  }
  else {
    // pull it out of the circular list
    if (first == least) {
      least = first->next;
    }

    first->previous->next = first->next;
    first->next->previous = first->previous;
    first = first->next;
  }



  length--;

  // if the queue is not empty, the first item better not be null, and visa versa
  ASSERT ( (length > 0 && first != NULL) || (length == 0 && first == NULL));

  return OK;
}

Status RequestQueue::verify () {
  int count = 0;
  Request *current = first;
  if (first == NULL) {
    return OK;
  }

  while (current->next != first) {
    ASSERT (current->action == READ || (current->action == WRITE));
    ASSERT (current->frameNumber >= 0 && current->frameNumber < Disk::diskFrameCount);
    ASSERT (current->data != NULL);
    //    ASSERT (current->data >= BufMgr::lowFrameAddress && current->data <=  BufMgr::highFrameAddress);
    count++;
    current = current->next;
  }

  ASSERT (current->action == READ || (current->action == WRITE));
  ASSERT (current->frameNumber >= 0 && current->frameNumber < Disk::diskFrameCount);
  ASSERT (current->data != NULL);
  //  ASSERT (current->data >=  BufMgr::lowFrameAddress && current->data <=  BufMgr::highFrameAddress);
  count++;


  ASSERT (count == length);

  return OK;
}

ostream& operator << (ostream& os, RequestQueue& queue) {
  if (queue.least != NULL) {
    Request *current = queue.least;
    bool seenLast = false;
    while (!seenLast || current != queue.least) {
      //      os << current->frameNumber;
      int offset = current->frameNumber % CFS::framesPerGroup; \
      int groupNo = current->frameNumber / CFS::framesPerGroup; \
      cout << "(" << groupNo << "," << offset << ") "; \


      if (current == queue.first)
	os << "* ";
      else
	os << " ";
      if (current->next == queue.least) {
	seenLast = true;
      }
      current = current->next;
    }
    os << endl;
  } else {
    os << "Empty.";
  }
  return os;
}

#endif


